class ModularOdmException(Exception):
    """ base class from which all exceptions raised by modularodm should inherit
    """
    pass


class QueryException(ModularOdmException):
    """ base class for exceptions raised from query parsing or execution"""
    pass


class KeyExistsException(QueryException):
    """Raised when a insert in an object with a primary key that already exists.
    """
    pass


class MultipleResultsFound(QueryException):
    """ Raised when multiple results match the passed query, and only a single
    object may be returned """
    pass


class NoResultsFound(QueryException):
    """ Raised when no results match the passed query, but one or more results
    must be returned. """
    pass


class ValidationError(ModularOdmException):
    """ Base class for exceptions raised during validation. Should not raised
    directly. """
    pass


class ValidationTypeError(ValidationError, TypeError):
    """ Raised during validation if explicit type check failed """
    pass


class ValidationValueError(ValidationError, ValueError):
    """ Raised during validation if the value of the input is unacceptable, but
     the type is correct """
    pass


class ImproperConfigurationError(ModularOdmException):
    """Raised if configuration options are not set correctly."""
    pass

class DatabaseError(ModularOdmException):
    '''Raised when execution of a database operation fails.'''
    pass



# -*- coding: utf-8 -*-

import weakref
import collections
import six
from werkzeug.local import LocalProxy

from modularodm.cache import Cache
from modularodm.writequeue import WriteQueue


# Dictionary of proxies, keyed on base schema, class variable label, and
# extension-specific key (e.g. a Flask request). The final key uses a weak
# reference to avoid memory leaks.
proxies = collections.defaultdict(
    lambda: collections.defaultdict(weakref.WeakKeyDictionary)
)

# Class variables on ``StoredObject`` that should be request-local under
# concurrent use
proxied_members = {
    '_cache': Cache,
    '_object_cache': Cache,
    'queue': WriteQueue,
}


def proxy_factory(BaseSchema, label, ProxiedClass, get_key):
    """Create a proxy to a class instance stored in ``proxies``.

    :param class BaseSchema: Base schema (e.g. ``StoredObject``)
    :param str label: Name of class variable to set
    :param class ProxiedClass: Class to get or create
    :param function get_key: Extension-specific key function; may return e.g.
        the current Flask request

    """
    def local():
        key = get_key()
        try:
            return proxies[BaseSchema][label][key]
        except KeyError:
            proxies[BaseSchema][label][key] = ProxiedClass()
            return proxies[BaseSchema][label][key]
    return LocalProxy(local)


def with_proxies(proxy_map, get_key):
    """Class decorator factory; adds proxy class variables to target class.

    :param dict proxy_map: Mapping between class variable labels and proxied
        classes
    :param function get_key: Extension-specific key function; may return e.g.
        the current Flask request

    """
    def wrapper(cls):
        for label, ProxiedClass in six.iteritems(proxy_map):
            proxy = proxy_factory(cls, label, ProxiedClass, get_key)
            setattr(cls, label, proxy)
        return cls
    return wrapper



import abc
import six
import time
import random
import itertools
from functools import wraps

from ..translators import DefaultTranslator
from modularodm.exceptions import KeyExistsException


class Logger(object):

    def __init__(self):

        self.listening = False
        self.events = []
        self.xtra = []

    def listen(self, xtra=None):

        self.xtra.append(xtra)

        if self.listening:
            return False

        self.listening = True
        self.events = []
        return True

    def record_event(self, event):

        if self.listening:
            self.events.append(event)

    def report(self, sort_func=None):

        out = {}

        if sort_func is None:
            sort_func = lambda e: e.func.__name__

        heard = sorted(self.events, key=sort_func)

        for key, group in itertools.groupby(heard, sort_func):
            group = list(group)
            num_events = len(group)
            total_time = sum([event.elapsed_time for event in group])
            out[key] = (num_events, total_time)

        return out

    def pop(self):

        self.xtra.pop()

    def clear(self):

        self.listening = False
        self.events = []


class LogEvent(object):

    def __init__(self, func, start_time, stop_time, xtra=None):

        self.func = func
        self.start_time = start_time
        self.stop_time = stop_time
        self.elapsed_time = stop_time - start_time
        self.xtra = xtra

    def __repr__(self):

        return 'LogEvent("{func}", {start_time}, {stop_time}, {xtra})'.format(
            **self.__dict__
        )


def logify(func):

    @wraps(func)
    def wrapped(this, *args, **kwargs):

        # Note: Copy value of `this.logger.listening` here in the event that
        # this value is changed externally during the decorated function call.
        # TODO: Verify that this produces valid output for concurrent requests
        listening = this.logger.listening

        if listening:
            start_time = time.time()

        out = func(this, *args, **kwargs)

        if listening:
            stop_time = time.time()
            # TODO: This is a temporary fix for a suspected concurrency issue.
            xtra = this.logger.xtra[-1] if this.logger.xtra else None
            this.logger.record_event(
                LogEvent(
                    func,
                    start_time,
                    stop_time,
                    xtra
                )
            )

        return out

    return wrapped


class StorageMeta(abc.ABCMeta):

    def __new__(mcs, name, bases, dct):

        # Decorate methods
        for key, value in dct.items():
            if hasattr(value, '__call__') \
                    and not isinstance(value, type) \
                    and not key.startswith('_'):
                dct[key] = logify(value)

        # Run super-metaclass __new__
        return super(StorageMeta, mcs).__new__(mcs, name, bases, dct)


@six.add_metaclass(StorageMeta)
class Storage(object):
    """Abstract base class for storage objects. Subclasses (e.g.
    :class:`~modularodm.storage.picklestorage.PickleStorage`,
    :class:`~modularodm.storage.mongostorage.MongoStorage`, etc.)
    must define insert, update, get, remove, flush, and find_all methods.
    """
    translator = DefaultTranslator()
    logger = Logger()

    def _ensure_index(self, key):
        pass

    # todo allow custom id generator
    # todo increment n on repeated failures
    def _generate_random_id(self, n=5):
        """Generated random alphanumeric key.

        :param n: Number of characters in random key
        """
        alphabet = '23456789abcdefghijkmnpqrstuvwxyz'
        return ''.join(random.sample(alphabet, n))

    def _optimistic_insert(self, primary_name, value, n=5):
        """Attempt to insert with randomly generated key until insert
        is successful.

        :param str primary_name: The name of the primary key.
        :param dict value: The dictionary representation of the record.
        :param n: Number of characters in random key
        """
        while True:
            try:
                key = self._generate_random_id(n)
                value[primary_name] = key
                self.insert(primary_name, key, value)
                break
            except KeyExistsException:
                pass
        return key

    @abc.abstractmethod
    def insert(self, primary_name, key, value):
        """Insert a new record.

        :param str primary_name: Name of primary key
        :param key: The value of the primary key
        :param dict value: The dictionary of attribute:value pairs
        """
        pass

    @abc.abstractmethod
    def update(self, query, data):
        """Update multiple records with new data.

        :param query: A query object.
        :param dict data: Dictionary of key:value pairs.
        """
        pass

    @abc.abstractmethod
    def get(self, primary_name, key):
        """Get a single record.

        :param str primary_name: The name of the primary key.
        :param key: The value of the primary key.
        """
        pass

    @abc.abstractmethod
    def remove(self, query=None):
        """Remove records.
        """
        pass

    @abc.abstractmethod
    def flush(self):
        """Flush the database."""
        pass

    @abc.abstractmethod
    def find_one(self, query=None, **kwargs):
        """ Gets a single object from the collection.

        If no matching documents are found, raises `NoResultsFound`.
        If >1 matching documents are found, raises `MultipleResultsFound`.

        :params: One or more `Query` or `QuerySet` objects may be passed

        :returns: The selected document
        """
        pass

    @abc.abstractmethod
    def find(self, query=None, **kwargs):
        """
        Return a generator of query results. Takes optional `by_pk` keyword
        argument; if true, return keys rather than
        values.

        :param query:

        :return: a generator of :class:`~.storedobject.StoredObject` instances
        """
        pass



# -*- coding: utf-8 -*-

from flask import request

from modularodm.storedobject import StoredObject

from .concurrency import with_proxies, proxied_members


# ``FlaskStoredObject`` should work when outside a request context, so we use
# a dummy key in this case. Because we can't create weak references to some
# types (str, NoneType, etc.), use an object as the dummy key.
class DummyRequest(object):
    pass
dummy_request = DummyRequest()


def get_flask_key():
    """Get the current Flask request; if not working in request context,
    return the dummy request.

    """
    try:
        return request._get_current_object()
    except RuntimeError:
        return dummy_request


@with_proxies(proxied_members, get_flask_key)
class FlaskStoredObject(StoredObject):
    pass



# -*- coding: utf-8 -*-

from modularodm import exceptions

from modularodm.fields.foreign import BaseForeignField
from .lists import ForeignList


class ForeignField(BaseForeignField):

    _list_class = ForeignList

    def __init__(self, *args, **kwargs):

        super(ForeignField, self).__init__(*args, **kwargs)

        self._backref_field_name = kwargs.get('backref', None)
        self._base_class_reference = args[0]
        self._base_class = None
        self._is_foreign = True
        self._is_abstract = False

    def get_foreign_object(self, value):
        return self.base_class.load(value)

    def to_storage(self, value, translator=None):

        if value is None:
            return value
        try:
            value_to_store = value._primary_key
        except AttributeError:
            value_to_store = value
        _foreign_pn = self.base_class._primary_name
        return self.base_class._fields[_foreign_pn].to_storage(value_to_store, translator)

    def from_storage(self, value, translator=None):

        if value is None:
            return None
        _foreign_pn = self.base_class._primary_name
        _foreign_pk = self.base_class._fields[_foreign_pn].from_storage(value, translator)
        return _foreign_pk

    def _to_primary_key(self, value):
        """
        Return primary key; if value is StoredObject, verify
        that it is loaded.

        """
        if value is None:
            return None
        if isinstance(value, self.base_class):
            if not value._is_loaded:
                raise exceptions.DatabaseError('Record must be loaded.')
            return value._primary_key

        return self.base_class._to_primary_key(value)
        # return self.base_class._check_pk_type(value)

    @property
    def mutable(self):
        return self.base_class._fields[self.base_class._primary_name].mutable

    @property
    def base_class(self):
        if self._base_class:
            return self._base_class
        if isinstance(self._base_class_reference, type):
            self._base_class = self._base_class_reference
        else:
            try:
                self._base_class = self._schema_class.get_collection(
                    self._base_class_reference
                )
            except KeyError:
                raise exceptions.ModularOdmException(
                    'Unknown schema "{0}"'.format(
                        self._base_class_reference
                    )
                )
        return self._base_class

    def __set__(self, instance, value, safe=False, literal=False):
        # if instance._detached:
        #     warnings.warn('Accessing a detached record.')
        value_to_set = value if literal else self._to_primary_key(value)
        super(ForeignField, self).__set__(
            instance,
            value_to_set,
            safe=safe
        )

    def __get__(self, instance, owner):
        # if instance._detached:
        #     warnings.warn('Accessing a detached record.')
        primary_key = super(ForeignField, self).__get__(instance, None)
        if primary_key is None:
            return
        return self.base_class.load(primary_key)



import weakref
import warnings
import copy
import six

from modularodm import exceptions
from modularodm.query.querydialect import DefaultQueryDialect as Q
from .lists import List


def print_arg(arg):
    if isinstance(arg, six.string_types):
        return '"' + arg + '"'
    return arg


class Field(object):

    default = None
    base_class = None
    _list_class = List
    mutable = False
    lazy_default = True
    _uniform_translator = True

    def __repr__(self):
        return '{cls}({kwargs})'.format(
            cls=self.__class__.__name__,
            kwargs=', '.join('{}={}'.format(key, print_arg(val)) for key, val in self._kwargs.items())
        )

    def subscribe(self, sender=None):
        pass

    def _to_comparable(self):
        return {
            k : v
            for k, v in self.__dict__.items()
            if k not in ['data', '_translators', '_schema_class']
        }

    def __eq__(self, other):
        return self._to_comparable() == other._to_comparable()

    def __ne__(self, other):
        return not self.__eq__(other)

    def _prepare_validators(self, _validate):

        if hasattr(_validate, '__iter__'):

            # List of callable validators
            validate = []
            for validator in _validate:
                if hasattr(validator, '__call__'):
                    validate.append(validator)
                else:
                    raise TypeError('Validator lists must be lists of callables.')

        elif hasattr(_validate, '__call__'):

            # Single callable validator
            validate = _validate

        elif type(_validate) == bool:

            # Boolean validator
            validate = _validate

        else:

            # Invalid validator type
            raise TypeError('Validators must be callables, lists of callables, or booleans.')

        return _validate, validate

    def __init__(self, *args, **kwargs):

        self._args = args
        self._kwargs = kwargs
        self._translators = {}

        # Pointer to containing ListField
        # Set in StoredObject.ObjectMeta
        self._list_container = None

        self.data = weakref.WeakKeyDictionary()

        self._validate, self.validate = \
            self._prepare_validators(kwargs.get('validate', False))

        self._default = kwargs.get('default', self.default)
        self._is_primary = kwargs.get('primary', False)
        self._list = kwargs.get('list', False)
        self._required = kwargs.get('required', False)
        self._unique = kwargs.get('unique', False)
        self._editable = kwargs.get('editable', True)
        self._index = kwargs.get('index', self._is_primary)
        self._is_foreign = False

        # Fields added by ``ObjectMeta``
        self._field_name = None

    def do_validate(self, value, obj):

        # Check if required
        if value is None:
            if getattr(self, '_required', None):
                raise exceptions.ValidationError('Value <{0}> is required.'.format(self._field_name))
            return True

        # Check if unique
        if value is not None and self._unique:
            unique_query = Q(self._field_name, 'eq', value)
            # If object has primary key, don't crash if unique value is
            # already associated with its key
            if obj._is_loaded:
                unique_query = unique_query & Q(obj._primary_name, 'ne', obj._primary_key)
            if obj.find(unique_query).limit(1).count():
                raise exceptions.ValidationValueError('Value must be unique')

        # Field-level validation
        cls = self.__class__
        if hasattr(cls, 'validate') and self.validate is not False:
            cls.validate(value)

        # Schema-level validation
        if self._validate and hasattr(self, 'validate'):
            if hasattr(self.validate, '__iter__'):
                for validator in self.validate:
                    validator(value)
            elif hasattr(self.validate, '__call__'):
                self.validate(value)

        # Success
        return True

    def _gen_default(self):
        if callable(self._default):
            return self._default()
        return copy.deepcopy(self._default)

    def _get_translate_func(self, translator, direction):
        try:
            return self._translators[(translator, direction)]
        except KeyError:
            method_name = '%s_%s' % (direction, self.data_type.__name__)
            default_name = '%s_default' % (direction,)
            try:
                method = getattr(translator, method_name)
            except AttributeError:
                method = getattr(translator, default_name)
            self._translators[(translator, direction)] = method
            return method

    def to_storage(self, value, translator=None):
        translator = translator or self._schema_class._translator
        if value is None:
            return translator.null_value
        method = self._get_translate_func(translator, 'to')
        value = value if method is None else method(value)
        if self.mutable:
            return copy.deepcopy(value)
        return value

    def from_storage(self, value, translator=None):
        translator = translator or self._schema_class._translator
        if value == translator.null_value:
            return None
        method = self._get_translate_func(translator, 'from')
        value = value if method is None else method(value)
        if self.mutable:
            return copy.deepcopy(value)
        return value

    def _pre_set(self, instance, safe=False):
        if not self._editable and not safe:
            raise AttributeError('Field cannot be edited.')
        if instance._detached:
            warnings.warn('Accessing a detached record.')

    def __set__(self, instance, value, safe=False, literal=False):
        self._pre_set(instance, safe=safe)
        if self.mutable:
            value = copy.deepcopy(value)
        self.data[instance] = value

    def _touch(self, instance):

        # Reload if dirty
        if instance._dirty:
            instance._dirty = False
            instance.reload()

        # Impute default and return
        try:
            self.data[instance]
        except KeyError:
            self.data[instance] = self._gen_default()

    def __get__(self, instance, owner, check_dirty=True):

        # Warn if detached
        if instance._detached:
            warnings.warn('Accessing a detached record.')

        # Reload if dirty
        self._touch(instance)

        # Impute default and return
        try:
            return self.data[instance]
        except KeyError:
            default = self._gen_default()
            self.data[instance] = default
            return default

    def _get_underlying_data(self, instance):
        """Return data from raw data store, rather than overridden
        __get__ methods. Should NOT be overwritten.
        """
        self._touch(instance)
        return self.data.get(instance, None)

    def __delete__(self, instance):
        self.data.pop(instance, None)



# -*- coding: utf-8 -*-

def set_nested(data, value, *keys):
    """Assign to a nested dictionary.

    :param dict data: Dictionary to mutate
    :param value: Value to set
    :param list *keys: List of nested keys

    >>> data = {}
    >>> set_nested(data, 'hi', 'k0', 'k1', 'k2')
    >>> data
    {'k0': {'k1': {'k2': 'hi'}}}

    """
    if len(keys) == 1:
        data[keys[0]] = value
    else:
        if keys[0] not in data:
            data[keys[0]] = {}
        set_nested(data[keys[0]], value, *keys[1:])


class Cache(object):
    """Simple container for storing cached data.

    """
    def __init__(self):
        self.data = {}

    @property
    def raw(self):
        return self.data

    def set(self, schema, key, value):
        set_nested(self.data, value, schema, key)

    def get(self, schema, key):
        try:
            return self.data[schema][key]
        except KeyError:
            return None

    def pop(self, schema, key):
        self.data[schema].pop(key, None)

    def clear(self):
        self.__init__()

    def clear_schema(self, schema):
        self.data.pop(schema, None)

    def __nonzero__(self):
        return bool(self.data)

    # Python 3
    __bool__ = __nonzero__



from io import BytesIO

from .picklestorage import PickleStorage

try:
    import cpickle as pickle
except ImportError:
    import pickle


class EphemeralStorage(PickleStorage):
    def __init__(self, *args, **kwargs):
        self.store = {}
        self.fp = BytesIO()

    def flush(self):
        pickle.dump(self.store, self.fp, -1)



from .field import Field

from .booleanfield import BooleanField
from .datetimefield import DateTimeField
from .dictionaryfield import DictionaryField
from .floatfield import FloatField
from .foreignfield import ForeignField
from .abstractforeignfield import AbstractForeignField
from .integerfield import IntegerField
from .listfield import ListField
from .objectidfield import ObjectIdField
from .stringfield import StringField

from .lists import List, ForeignList, AbstractForeignList



# -*- coding: utf-8 -*_

import re
import pymongo

from .base import Storage
from ..query.queryset import BaseQuerySet
from ..query.query import QueryGroup
from ..query.query import RawQuery
from modularodm.exceptions import (
    KeyExistsException,
    MultipleResultsFound,
    NoResultsFound,
)


# From mongoengine.queryset.transform
COMPARISON_OPERATORS = ('ne', 'gt', 'gte', 'lt', 'lte', 'in', 'nin', 'mod',
                        'all', 'size', 'exists', 'not', 'elemMatch')
# GEO_OPERATORS        = ('within_distance', 'within_spherical_distance',
#                         'within_box', 'within_polygon', 'near', 'near_sphere',
#                         'max_distance', 'geo_within', 'geo_within_box',
#                         'geo_within_polygon', 'geo_within_center',
#                         'geo_within_sphere', 'geo_intersects')
STRING_OPERATORS     = ('contains', 'icontains', 'startswith',
                        'istartswith', 'endswith', 'iendswith',
                        'exact', 'iexact')
# CUSTOM_OPERATORS     = ('match',)
# MATCH_OPERATORS      = (COMPARISON_OPERATORS + GEO_OPERATORS +
#                         STRING_OPERATORS + CUSTOM_OPERATORS)

# UPDATE_OPERATORS     = ('set', 'unset', 'inc', 'dec', 'pop', 'push',
#                         'push_all', 'pull', 'pull_all', 'add_to_set',
#                         'set_on_insert')


# Adapted from mongoengine.fields
def prepare_query_value(op, value):

    if op.lstrip('i') in ('startswith', 'endswith', 'contains', 'exact'):
        flags = 0
        if op.startswith('i'):
            flags = re.IGNORECASE
            op = op.lstrip('i')

        regex = r'%s'
        if op == 'startswith':
            regex = r'^%s'
        elif op == 'endswith':
            regex = r'%s$'
        elif op == 'exact':
            regex = r'^%s$'

        # escape unsafe characters which could lead to a re.error
        value = re.escape(value)
        value = re.compile(regex % value, flags)

    return value


# TODO: Test me
def translate_query(query=None, mongo_query=None):
    """

    """
    mongo_query = mongo_query or {}

    if isinstance(query, RawQuery):
        attribute, operator, argument = \
            query.attribute, query.operator, query.argument

        if operator == 'eq':
            mongo_query[attribute] = argument

        elif operator in COMPARISON_OPERATORS:
            mongo_operator = '$' + operator
            if attribute not in mongo_query:
                mongo_query[attribute] = {}
            mongo_query[attribute][mongo_operator] = argument

        elif operator in STRING_OPERATORS:
            mongo_operator = '$regex'
            mongo_regex = prepare_query_value(operator, argument)
            if attribute not in mongo_query:
                mongo_query[attribute] = {}
            mongo_query[attribute][mongo_operator] = mongo_regex

    elif isinstance(query, QueryGroup):

        if query.operator == 'and':
            return {'$and': [translate_query(node) for node in query.nodes]}

        elif query.operator == 'or':
            return {'$or' : [translate_query(node) for node in query.nodes]}

        elif query.operator == 'not':
            # Hack: A nor A == not A
            subquery = translate_query(query.nodes[0])
            return {'$nor' : [subquery, subquery]}

        else:
            raise ValueError('QueryGroup operator must be <and>, <or>, or <not>.')

    elif query is None:
        return {}

    else:
        raise TypeError('Query must be a QueryGroup or Query object.')

    return mongo_query


class MongoQuerySet(BaseQuerySet):

    _NEGATIVE_INDEXING = True

    def __init__(self, schema, cursor):
        super(MongoQuerySet, self).__init__(schema)
        self.data = cursor
        self._order = [('_id', 1)]  # Default sorting

    def _do_getitem(self, index, raw=False):
        if isinstance(index, slice):
            return MongoQuerySet(self.schema, self.data.clone()[index])
        if index < 0:
            clone = self.data.clone().sort([(o[0], o[1] * -1) for o in self._order])
            result = clone[(index * -1) - 1]
        else:
            result = self.data[index]
        if raw:
            return result[self.primary]
        return self.schema.load(data=result)

    def __iter__(self, raw=False):
        cursor = self.data.clone()
        if raw:
            return [each[self.primary] for each in cursor]
        return (self.schema.load(data=each) for each in cursor)

    def __len__(self):
        return self.data.count(with_limit_and_skip=True)

    count = __len__

    def get_key(self, index):
        return self.__getitem__(index, raw=True)

    def get_keys(self):
        return list(self.__iter__(raw=True))

    def sort(self, *keys):
        sort_key = []

        for key in keys:

            if key.startswith('-'):
                key = key.lstrip('-')
                sign = pymongo.DESCENDING
            else:
                sign = pymongo.ASCENDING

            sort_key.append((key, sign))

        self._order = sort_key
        self.data = self.data.sort(sort_key)
        return self

    def offset(self, n):
        self.data = self.data.skip(n)
        return self

    def limit(self, n):
        self.data = self.data.limit(n)
        return self


class MongoStorage(Storage):
    """Wrap a MongoDB collection. Note: `store` is a property instead of an
    attribute to handle passing `db` as a proxy.

    :param Database db:
    :param str collection:
    """
    QuerySet = MongoQuerySet

    def __init__(self, db, collection):
        self.db = db
        self.collection = collection

    @property
    def store(self):
        return self.db[self.collection]

    def _ensure_index(self, key):
        self.store.ensure_index(key)

    def find(self, query=None, **kwargs):
        mongo_query = translate_query(query)
        return self.store.find(mongo_query)

    def find_one(self, query=None, **kwargs):
        mongo_query = translate_query(query)
        matches = self.store.find(mongo_query).limit(2)

        if matches.count() == 1:
            return matches[0]

        if matches.count() == 0:
            raise NoResultsFound()

        raise MultipleResultsFound(
            'Query for find_one must return exactly one result; '
            'returned {0}'.format(matches.count())
        )

    def get(self, primary_name, key):
        return self.store.find_one({primary_name : key})

    def insert(self, primary_name, key, value):
        if primary_name not in value:
            value = value.copy()
            value[primary_name] = key
        try:
            self.store.insert(value)
        except pymongo.errors.DuplicateKeyError:
            raise KeyExistsException

    def update(self, query, data):
        mongo_query = translate_query(query)

        # Field "_id" shouldn't appear in both search and update queries; else
        # MongoDB will raise a "Mod on _id not allowed" error
        if '_id' in mongo_query:
            update_data = {k: v for k, v in data.items() if k != '_id'}
        else:
            update_data = data
        update_query = {'$set': update_data}

        self.store.update(
            mongo_query,
            update_query,
            upsert=False,
            multi=True,
        )

    def remove(self, query=None):
        mongo_query = translate_query(query)
        self.store.remove(mongo_query)

    def flush(self):
        pass



from dateutil import parser as dateparser
from bson import ObjectId

class DefaultTranslator(object):

    null_value = None

    to_default = None
    from_default = None

class JSONTranslator(DefaultTranslator):

    def to_datetime(self, value):
        return str(value)

    def from_datetime(self, value):
        return dateparser.parse(value)

    def to_ObjectId(self, value):
        return str(value)

    def from_ObjectId(self, value):
        return ObjectId(value)

class StringTranslator(JSONTranslator):

    null_value = 'none'

    def to_default(self, value):
        return str(value)

    def from_int(self, value):
        return int(value)

    def from_float(self, value):
        return float(value)

    def from_bool(self, value):
        return bool(value)



import collections

def freeze(value):
    """ Cast value to its frozen counterpart. """
    if isinstance(value, list):
        return FrozenList(*value)
    if isinstance(value, dict):
        return FrozenDict(**value)
    return value

def thaw(value):
    if isinstance(value, FrozenList):
        return value.thaw()
    if isinstance(value, FrozenDict):
        return value.thaw()
    return value

class FrozenDict(collections.Mapping):
    """ Immutable dictionary. """
    def __init__(self, **kwargs):
        self.__data = {key : freeze(value) for key, value in kwargs.items()}

    def thaw(self):
        return {key : thaw(value) for key, value in self.__data.items()}

    def __eq__(self, other):
        if not isinstance(other, FrozenDict):
            return self.thaw() == other
        return super(FrozenDict, self).__eq__(self, other)

    def __getitem__(self, item):
        return self.__data[item]

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        return repr(self.__data)

class FrozenList(collections.Sequence):
    """ Immutable list. """
    def __init__(self, *args):
        self.__data = [freeze(value) for value in args]

    def thaw(self):
        return [thaw(value) for value in self.__data]

    def __eq__(self, other):
        if not isinstance(other, FrozenList):
            return self.thaw() == other
        return super(FrozenList, self).__eq__(self, other)

    def __getitem__(self, item):
        return self.__data[item]

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        return repr(self.__data)


# -*- coding: utf-8 -*-

import datetime

from modularodm import signals
from modularodm.fields import Field
from modularodm.validators import validate_datetime


DEFAULT_NOW = datetime.datetime.utcnow


def default_or_callable(value):
    if value is True:
        return DEFAULT_NOW
    if callable(value):
        return value
    raise ValueError('Value must be True or callable')


class DateTimeField(Field):

    validate = validate_datetime
    data_type = datetime.datetime
    mutable = True

    def __init__(self, *args, **kwargs):

        super(DateTimeField, self).__init__(*args, **kwargs)

        auto_now = kwargs.pop('auto_now', False)
        auto_now_add = kwargs.pop('auto_now_add', False)
        if auto_now and auto_now_add:
            raise ValueError('Cannot use auto_now and auto_now_add on the '
                             'same field.')

        #
        if (auto_now or auto_now_add) and 'editable' not in kwargs:
            self._editable = False
            self.lazy_default = False

        #
        if auto_now:
            self._auto_now = default_or_callable(auto_now)
        elif auto_now_add:
            self._default = default_or_callable(auto_now_add)

    def subscribe(self, sender=None):
        self.update_auto_now = signals.before_save.connect(
            self.update_auto_now,
            sender=sender,
        )

    def update_auto_now(self, cls, instance):
        if getattr(self, '_auto_now', None):
            self.data[instance] = self._auto_now()



# -*- coding utf-8 -*-

import abc
import six

from modularodm import signals
from modularodm.fields import Field


@six.add_metaclass(abc.ABCMeta)
class BaseForeignField(Field):

    @abc.abstractmethod
    def get_foreign_object(self, value):
        pass

    def update_backrefs(self, instance, cached_value, current_value):

        if self._backref_field_name is None:
            return

        if cached_value:
            cached_object = self.get_foreign_object(cached_value)
            if cached_object:
                cached_object._remove_backref(
                    self._backref_field_name,
                    instance,
                    self._field_name
                )

        if current_value:
            current_value._set_backref(
                self._backref_field_name,
                self._field_name,
                instance
            )

    def update_backrefs_callback(self, cls, instance, fields_changed, cached_data):

        if self._field_name not in fields_changed:
            return

        cached_value = cached_data.get(self._field_name)
        current_value = getattr(instance, self._field_name)

        self.update_backrefs(instance, cached_value, current_value)

    def subscribe(self, sender=None):
        self.update_backrefs_callback = signals.save.connect(
            self.update_backrefs_callback,
            sender=sender,
        )



# -*- coding: utf-8 -*-

import copy

import six

from modularodm import signals
from ..fields import Field
from ..validators import validate_list


class ListField(Field):

    validate = validate_list

    def __init__(self, field_instance, **kwargs):

        super(ListField, self).__init__(**kwargs)

        self._list_validate, self.list_validate = self._prepare_validators(kwargs.get('list_validate', False))

        # ListField is a list of the following (e.g., ForeignFields)
        self._field_instance = field_instance
        self._is_foreign = field_instance._is_foreign
        self._is_abstract = getattr(field_instance, '_is_abstract', False)
        self._uniform_translator = field_instance._uniform_translator

        # Descriptor data is this type of list
        self._list_class = self._field_instance._list_class

        # Descriptor data is this type of list object, instantiated as our
        # default
        if self._default:
            default = self._default() if six.callable(self._default) else self._default
            if (
                not hasattr(default, '__iter__')
                or isinstance(default, dict)
                or isinstance(default, six.string_types)
            ):
                raise TypeError(
                    'Default value for list fields must be a list; received {0}'.format(
                        type(self._default)
                    )
                )
        else:
            default = None

        #if (self._default
        #    and not hasattr(self._default, '__iter__')
        #    or isinstance(self._default, dict)):
        #    raise TypeError(
        #        'Default value for list fields must be a list; received {0}'.format(
        #            type(self._default)
        #        )
        #    )

        # Default is a callable that returns an empty instance of the list class
        # Avoids the need to deepcopy default values for lists, which will break
        # e.g. when validators contain (un-copyable) regular expressions.
        self._default = lambda: self._list_class(default, base_class=self._field_instance.base_class)

        # Fields added by ``ObjectMeta``
        self._field_name = None

    def subscribe(self, sender=None):
        self.update_backrefs_callback = signals.save.connect(
            self.update_backrefs_callback,
            sender=sender,
        )

    def __set__(self, instance, value, safe=False, literal=False):
        self._pre_set(instance, safe=safe)
        # if isinstance(value, self._default.__class__):
        #     self.data[instance] = value
        if hasattr(value, '__iter__') and not isinstance(value, six.string_types):
            if literal:
                self.data[instance] = self._list_class(value, base_class=self._field_instance.base_class, literal=True)
            else:
                self.data[instance] = self._list_class(base_class=self._field_instance.base_class)
                self.data[instance].extend(value)
        else:
            self.data[instance] = value

    def do_validate(self, value, obj):

        inst = self._field_instance
        if getattr(inst, 'validate', False) or inst._unique or inst._required:
            # Child-level validation
            for part in value:
                self._field_instance.do_validate(part, obj)

        # Field-level list validation
        if hasattr(self.__class__, 'validate'):
            self.__class__.validate(value)

        # Schema-level list validation
        if self._list_validate:
            if hasattr(self.list_validate, '__iter__'):
                for validator in self.list_validate:
                    validator(value)
            elif hasattr(self.list_validate, '__call__'):
                self.list_validate(value)

        # Success
        return True

    def _get_translate_func(self, translator, direction):
        try:
            return self._translators[(translator, direction)]
        except KeyError:
            if self._is_foreign:
                base_class = self._field_instance.base_class
                primary_field = base_class._fields[base_class._primary_name]
                method = primary_field._get_translate_func(translator, direction)
            else:
                method = self._field_instance._get_translate_func(translator, direction)
            self._translators[(translator, direction)] = method
            return method

    def to_storage(self, value, translator=None):
        translator = translator or self._schema_class._translator
        if value:
            if hasattr(value, '_to_data'):
                value = value._to_data()
            if self._uniform_translator:
                method = self._get_translate_func(translator, 'to')
                if method is not None or translator.null_value is not None:
                    value = [
                        translator.null_value if item is None
                        else
                        item if method is None
                        else
                        method(item)
                        for item in value
                    ]
                if self._field_instance.mutable:
                    return copy.deepcopy(value)
                return copy.copy(value)
            else:
                return [
                    self._field_instance.to_storage(item)
                    for item in value
                ]
        return []

    def from_storage(self, value, translator=None):
        translator = translator or self._schema_class._translator
        if value:
            if self._uniform_translator:
                method = self._get_translate_func(translator, 'from')
                if method is not None or translator.null_value is not None:
                    value = [
                        None if item is translator.null_value
                        else
                        item if method is None
                        else
                        method(item)
                        for item in value
                    ]
                if self._field_instance.mutable:
                    return copy.deepcopy(value)
                return copy.copy(value)
            else:
                return [
                    self._field_instance.from_storage(item)
                    for item in value
                ]
        return []

    def update_backrefs(self, instance, cached_value, current_value):
        if self._field_instance._backref_field_name is None:
            return

        for item in current_value:
            if self._field_instance.to_storage(item) not in cached_value:
                self._field_instance.update_backrefs(instance, None, item)

        for item in cached_value:
            if self._field_instance.from_storage(item) not in current_value:
                self._field_instance.update_backrefs(instance, item, None)

    def update_backrefs_callback(self, cls, instance, fields_changed, cached_data):

        if not hasattr(self._field_instance, 'update_backrefs'):
            return

        if self._field_name not in fields_changed:
            return

        cached_value = cached_data.get(self._field_name, [])
        current_value = getattr(instance, self._field_name, [])

        self.update_backrefs(instance, cached_value, current_value)

    @property
    def base_class(self):
        if self._field_instance is None:
            return
        if not hasattr(self._field_instance, 'base_class'):
            return
        return self._field_instance.base_class



# -*- coding: utf-8 -*-

import abc
import six

from modularodm.query.querydialect import DefaultQueryDialect as Q


@six.add_metaclass(abc.ABCMeta)
class List(list):

    def __init__(self, value=None, literal=False, **kwargs):

        value = value or []
        self._base_class = kwargs.get('base_class', None)

        if literal:
            super(List, self).__init__(value)
        else:
            super(List, self).__init__()
            self.extend(value)


class BaseForeignList(List):

    @abc.abstractmethod
    def _to_primary_keys(self):
        pass

    @abc.abstractmethod
    def _from_value(self, value):
        pass

    def _to_data(self):
        return list(super(BaseForeignList, self).__iter__())

    def __iter__(self):
        if self:
            return (self[idx] for idx in range(len(self)))
        return iter([])

    def __setitem__(self, key, value):
        super(BaseForeignList, self).__setitem__(key, self._from_value(value))

    def insert(self, index, value):
        super(BaseForeignList, self).insert(index, self._from_value(value))

    def append(self, value):
        super(BaseForeignList, self).append(self._from_value(value))

    def extend(self, iterable):
        for item in iterable:
            self.append(item)

    def remove(self, value):
        super(BaseForeignList, self).remove(self._from_value(value))


class ForeignList(BaseForeignList):

    def _from_value(self, value):
        return self._base_class._to_primary_key(value)

    def _to_primary_keys(self):
        return self._to_data()

    def __reversed__(self):
        return ForeignList(
            super(ForeignList, self).__reversed__(),
            base_class=self._base_class
        )

    def __getitem__(self, item):
        result = super(ForeignList, self).__getitem__(item)
        if isinstance(item, slice):
            return ForeignList(result, base_class=self._base_class)
        return self._base_class.load(result)

    def __getslice__(self, i, j):
        result = super(ForeignList, self).__getslice__(i, j)
        return ForeignList(result, base_class=self._base_class)

    def __contains__(self, item):
        keys = self._to_primary_keys()
        if isinstance(item, self._base_class):
            return item._primary_key in keys
        if isinstance(item, self._base_class._primary_type):
            return item in keys
        return False

    def index(self, value, start=None, stop=None):
        start = 0 if start is None else start
        stop = len(self) if stop is None else stop
        keys = self._to_primary_keys()
        if isinstance(value, self._base_class):
            return keys.index(value._primary_key, start, stop)
        if isinstance(value, self._base_class._primary_type):
            return keys.index(value, start, stop)
        raise ValueError('{0} is not in list'.format(value))

    def find(self, query=None):
        combined_query = Q(
            self._base_class._primary_name,
            'in',
            self._to_primary_keys()
        )
        if query is not None:
            combined_query = combined_query & query
        return self._base_class.find(combined_query)


class AbstractForeignList(BaseForeignList):

    def _from_value(self, value):
        if hasattr(value, '_primary_key'):
            return (
                value._primary_key,
                value._name
            )
        return value

    def _to_primary_keys(self):
        return [
            item[0]
            for item in self._to_data()
        ]

    def __reversed__(self):
        return AbstractForeignList(
            super(AbstractForeignList, self).__reversed__()
        )

    def get_foreign_object(self, value):
        from modularodm import StoredObject
        return StoredObject.get_collection(value[1])\
            .load(value[0])

    def __getitem__(self, item):
        result = super(AbstractForeignList, self).__getitem__(item)
        if isinstance(item, slice):
            return AbstractForeignList(result)
        return self.get_foreign_object(result)

    def __getslice__(self, i, j):
        result = super(AbstractForeignList, self).__getslice__(i, j)
        return AbstractForeignList(result)

    def __contains__(self, item):
        keys = self._to_primary_keys()
        if hasattr(item, '_primary_key'):
            return item._primary_key in keys
        elif isinstance(item, tuple):
            return item[0] in keys
        return item in keys

    def index(self, value, start=None, stop=None):
        start = 0 if start is None else start
        stop = len(self) if stop is None else stop
        keys = self._to_primary_keys()
        if hasattr(value, '_primary_key'):
            return keys.index(value._primary_key, start, stop)
        elif isinstance(value, tuple):
            return keys.index(value[0], start, stop)
        else:
            try:
                return keys.index(value, start, stop)
            except ValueError:
                raise ValueError('{0} not in list'.format(value))



from ..fields import Field


class DictionaryField(Field):

    data_type = dict
    mutable = True

    def __init__(self, *args, **kwargs):
        super(DictionaryField, self).__init__(*args, **kwargs)
        self._default = kwargs.get('default', {})



# -*- coding utf-8 -*-

import os
import copy

import six

from .base import Storage
from ..query.queryset import BaseQuerySet
from ..query.query import QueryGroup
from ..query.query import RawQuery

from modularodm.utils import DirtyField
from modularodm.exceptions import (
    KeyExistsException,
    MultipleResultsFound,
    NoResultsFound,
)

try:
    import cpickle as pickle
except ImportError:
    import pickle


def _eq(data, test):
    if isinstance(data, list):
        return test in data
    return data == test

operators = {

    'eq':   _eq,

    'ne':   lambda data, test: data != test,
    'gt':   lambda data, test: data > test,
    'gte':  lambda data, test: data >= test,
    'lt':   lambda data, test: data < test,
    'lte':  lambda data, test: data <= test,
    'in':   lambda data, test: data in test,
    'nin':  lambda data, test: data not in test,

    'startswith':  lambda data, test: data.startswith(test),
    'endswith':    lambda data, test: data.endswith(test),
    'contains':    lambda data, test: test in data,
    'icontains':   lambda data, test: test.lower() in data.lower(),

}


class PickleQuerySet(BaseQuerySet):

    _sort = DirtyField(None)
    _offset = DirtyField(None)
    _limit = DirtyField(None)

    def __init__(self, schema, data):

        super(PickleQuerySet, self).__init__(schema)

        self._data = list(data)
        self._dirty = True

        self.data = []

    def _eval(self):

        if self._dirty:

            self.data = self._data[:]

            if self._sort is not None:

                for key in self._sort[::-1]:

                    if key.startswith('-'):
                        reverse = True
                        key = key.lstrip('-')
                    else:
                        reverse = False

                    self.data = sorted(
                        self.data,
                        key=lambda record: record[key],
                        reverse=reverse
                    )

            if self._offset is not None:
                self.data = self.data[self._offset:]

            if self._limit is not None:
                self.data = self.data[:self._limit]

            self._dirty = False

        return self

    def _do_getitem(self, index, raw=False):
        self._eval()
        if isinstance(index, slice):
            return PickleQuerySet(self.schema, self.data[index])
        key = self.data[index][self.primary]
        result = self.data[index]
        if raw:
            return result[self.primary]
        return self.schema.load(data=result)

    def __iter__(self, raw=False):
        self._eval()
        if raw:
            return [each[self.primary] for each in self.data]
        return (self.schema.load(data=each) for each in self.data)

    def __len__(self):
        self._eval()
        return len(self.data)

    count = __len__

    def get_key(self, index):
        return self.__getitem__(index, raw=True)

    def get_keys(self):
        return list(self.__iter__(raw=True))

    def sort(self, *keys):
        """ Iteratively sort data by keys in reverse order. """
        self._sort = keys
        return self

    def offset(self, n):
        self._offset = n
        return self

    def limit(self, n):
        self._limit = n
        return self


class PickleStorage(Storage):
    """ Storage backend using pickle. """

    QuerySet = PickleQuerySet

    def __init__(self, collection_name,  prefix='db_', ext='pkl'):
        """Build pickle file name and load data if exists.

        :param collection_name: Collection name
        :param prefix: File prefix.
        :param ext: File extension.

        """
        # Build filename
        filename = collection_name + '.' + ext
        if prefix:
            self.filename = prefix + filename
        else:
            self.filename = filename

        # Initialize empty store
        self.store = {}

        # Load file if exists
        if os.path.exists(self.filename):
            with open(self.filename, 'rb') as fp:
                data = fp.read()
                self.store = pickle.loads(data)

    def _delete_file(self):
        try:
            os.remove(self.filename)
        except OSError:
            pass

    def insert(self, primary_name, key, value):
        if key not in self.store:
            self.store[key] = copy.deepcopy(value)
            self.flush()
        else:
            msg = 'Key ({key}) already exists'.format(key=key)
            raise KeyExistsException(msg)

    def update(self, query, data):
        data = copy.deepcopy(data)
        for pk in self.find(query, by_pk=True):
            for key, value in data.items():
                self.store[pk][key] = value

    def get(self, primary_name, key):
        data = self.store.get(key)
        if data is not None:
            return copy.deepcopy(data)

    def _remove_by_pk(self, key, flush=True):
        """Retrieve value from store.

        :param key: Key

        """
        try:
            del self.store[key]
        except Exception as error:
            pass
        if flush:
            self.flush()

    def remove(self, query=None):
        for key in self.find(query, by_pk=True):
            self._remove_by_pk(key, flush=False)
        self.flush()

    def flush(self):
        with open(self.filename, 'wb') as fp:
            pickle.dump(self.store, fp, -1)

    def find_one(self, query=None, **kwargs):
        results = list(self.find(query))
        if len(results) == 1:
            return results[0]
        elif len(results) == 0:
            raise NoResultsFound()
        else:
            raise MultipleResultsFound(
                'Query for find_one must return exactly one result; '
                'returned {0}'.format(len(results))
            )

    def _match(self, value, query):

        if isinstance(query, QueryGroup):

            matches = [self._match(value, node) for node in query.nodes]

            if query.operator == 'and':
                return all(matches)
            elif query.operator == 'or':
                return any(matches)
            elif query.operator == 'not':
                return not any(matches)
            else:
                raise ValueError('QueryGroup operator must be <and>, <or>, or <not>.')

        elif isinstance(query, RawQuery):
            attribute, operator, argument = \
                query.attribute, query.operator, query.argument

            return operators[operator](value[attribute], argument)

        else:
            raise TypeError('Query must be a QueryGroup or Query object.')

    def find(self, query=None, **kwargs):
        if query is None:
            for key, value in six.iteritems(self.store):
                yield value
        else:
            # TODO: Making this a generator breaks it, since it can change
            for key, value in list(six.iteritems(self.store)):
                if self._match(value, query):
                    if kwargs.get('by_pk'):
                        yield key
                    else:
                        yield value



# -*- coding: utf-8 -*-

import logging
import collections


logger = logging.getLogger(__name__)


class WriteAction(object):

    def __init__(self, method, *args, **kwargs):
        if not callable(method):
            raise ValueError('Argument `method` must be callable')
        self.method = method
        # Note: `args` and `kwargs` must not be mutated after an action is
        # enqueued and before it is committed, else awful things can happen
        self.args = args
        self.kwargs = kwargs

    def execute(self):
        return self.method(*self.args, **self.kwargs)

    def __repr__(self):
        return '{0}(*{1}, **{2})'.format(
            self.method.__name__,
            self.args,
            self.kwargs
        )


class WriteQueue(object):

    def __init__(self):
        self.active = False
        self.actions = collections.deque()

    def start(self):
        if self.active:
            logger.warn('Already working in a write queue. Further writes '
                        'will be appended to the current queue.')
        self.active = True

    def push(self, action):
        if not self.active:
            raise ValueError('Cannot push unless queue is active')
        if not isinstance(action, WriteAction):
            raise TypeError('Argument `action` must be instance '
                            'of `WriteAction`')
        self.actions.append(action)

    def commit(self):
        if not self.active:
            raise ValueError('Cannot commit unless queue is active')
        results = []
        while self.actions:
            action = self.actions.popleft()
            results.append(action.execute())
        return results

    def clear(self):
        self.active = False
        self.actions = collections.deque()

    def __nonzero__(self):
        return bool(self.actions)

    # Python 3
    __bool__ = __nonzero__


class QueueContext(object):

    def __init__(self, BaseSchema):
        self.BaseSchema = BaseSchema

    def __enter__(self):
        self.BaseSchema.start_queue()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.BaseSchema.commit_queue()



# -*- coding: utf-8 -*-

import weakref


# TODO: Test me @jmcarp
class CallbackField(object):

    def __init__(self, default, callback):
        self.data = weakref.WeakKeyDictionary()
        self.default = default
        self.callback = callback

    def __get__(self, instance, owner):
        try:
            return self.data[instance]
        except KeyError:
            return self.default

    def __set__(self, instance, value):
        current = self.__get__(instance, None)
        self.data[instance] = value
        self.callback(instance, value, current)


def set_dirty_factory(field='_dirty'):
    def set_dirty(instance, new, current):
        if new != current:
            setattr(instance, field, True)
    return set_dirty


class DirtyField(CallbackField):

    def __init__(self, default, field='_dirty'):
        super(DirtyField, self).__init__(default, set_dirty_factory(field))




from .query import RawQuery

class QueryDialect(object):

    pass

class DefaultQueryDialect(QueryDialect, RawQuery):

    pass

class DictQueryDialect(QueryDialect):

    pass

class DunderQueryDialect(QueryDialect):

    pass

# '''
# __Q(foo='bar', baz__startswith='fez')
# __Q(__or(foo='bar', baz__startswith='fez'))
# __Q(__and(__or(foo='bar', baz__startswith='fez'), qux__ne='mom'))
# '''


# -*- coding: utf-8 -*-

from modularodm.fields.foreign import BaseForeignField
from .lists import AbstractForeignList


class AbstractForeignField(BaseForeignField):

    _list_class = AbstractForeignList
    _is_foreign = True
    _uniform_translator = False

    def __init__(self, *args, **kwargs):
        super(AbstractForeignField, self).__init__(*args, **kwargs)
        self._backref_field_name = kwargs.get('backref', None)
        self._is_foreign = True
        self._is_abstract = True

    def get_schema_class(self, schema):
        return self._schema_class.get_collection(schema)

    def get_primary_field(self, schema):
        schema_class = self.get_schema_class(schema)
        return schema_class._fields[schema_class._primary_name]

    def get_foreign_object(self, value):
        return self.get_schema_class(value[1])\
            .load(value[0])

    def to_storage(self, value, translator=None):

        if value is None:
            return None
        if not hasattr(value, '__iter__'):
            value = (value._primary_key, value._name)
        return (
            self.get_primary_field(value[1])\
                .to_storage(value[0], translator),
            value[1]
        )

    def from_storage(self, value, translator=None):

        if value is None:
            return None
        return (
            self.get_primary_field(value[1])\
                .from_storage(value[0], translator),
            value[1]
        )

    def _to_primary_key(self, value):

        if value is None:
            return None
        if hasattr(value, '_primary_key'):
            return value._primary_key

    def __set__(self, instance, value, safe=False, literal=False):
        if hasattr(value, '_primary_key'):
            value = (
                value._primary_key,
                value._name
            )
        elif isinstance(value, tuple) or isinstance(value, list):
            if len(value) != 2:
                raise ValueError('Value must have length 2')
        elif value is not None:
            raise TypeError('Value must be StoredObject, tuple, or None')
        super(AbstractForeignField, self).__set__(
            instance, value, safe=safe, literal=literal
        )

    def __get__(self, instance, owner, check_dirty=True):
        value = super(AbstractForeignField, self).__get__(
            instance, None, check_dirty
        )
        if value is None:
            return None
        return self.get_foreign_object(value)



# -*- coding: utf-8 -*-

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class BaseQuerySet(object):

    _NEGATIVE_INDEXING = False

    def __init__(self, schema, data=None):

        self.schema = schema
        self.primary = schema._primary_name
        self.data = data

    def __getitem__(self, index):
        if isinstance(index, slice):
            if index.step:
                raise IndexError('Slice steps not supported')
            if (index.start is not None and index.start < 0) or (index.stop is not None and index.stop < 0):
                raise IndexError('Negative indexing not supported')
            if index.stop is not None and index.stop < index.start:
                raise IndexError('Stop index must be greater than start index')
        elif not self.__class__._NEGATIVE_INDEXING and index < 0:
            raise IndexError('Negative indexing not supported')
        return self._do_getitem(index)

    @abc.abstractmethod
    def _do_getitem(self, index):
        pass

    @abc.abstractmethod
    def __iter__(self):
        pass

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def count(self):
        pass

    @abc.abstractmethod
    def sort(self, *keys):
        pass

    @abc.abstractmethod
    def offset(self, n):
        pass

    @abc.abstractmethod
    def limit(self, n):
        pass



# -*- coding: utf-8 -*-
import re
import sys
import subprocess

import pip
from setuptools import setup, find_packages


def parse_requirements(requirements):
    with open(requirements) as f:
        return [
            l.strip('\n') for l
            in f if l.strip('\n') and not l.startswith('#')
        ]


def find_version(fname):
    '''Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.
    '''
    version = ''
    with open(fname, 'r') as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


__version__ = find_version("modularodm/__init__.py")

PUBLISH_CMD = "python setup.py register sdist bdist_wheel upload"
TEST_PUBLISH_CMD = 'python setup.py register -r test sdist bdist_wheel upload -r test'
TEST_CMD = 'nosetests'

if 'publish' in sys.argv:
    try:
        __import__('wheel')
    except ImportError:
        print("wheel required. Run `pip install wheel`.")
        sys.exit(1)
    status = subprocess.call(PUBLISH_CMD, shell=True)
    sys.exit(status)

if 'publish_test' in sys.argv:
    try:
        __import__('wheel')
    except ImportError:
        print("wheel required. Run `pip install wheel`.")
        sys.exit(1)
    status = subprocess.call(TEST_PUBLISH_CMD, shell=True)
    sys.exit()

if 'run_tests' in sys.argv:
    try:
        __import__('nose')
    except ImportError:
        print('nose required. Run `pip install nose`.')
        sys.exit(1)

    status = subprocess.call(TEST_CMD, shell=True)
    sys.exit(status)


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content

setup(
    name='modular-odm',
    version=__version__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ],
    url="https://github.com/CenterForOpenScience/modular-odm",
    author='Center for Open Science',
    author_email='contact@centerforopenscience.org',
    zip_safe=False,
    description='A Pythonic Object Data Manager',
    long_description=read("README.rst"),
    packages=find_packages(exclude=("test*",)),
    install_requires=parse_requirements('requirements.txt'),
    tests_require=["nose"],
    keywords=["odm", "nosql", "mongo", "mongodb"],
)



# -*- coding: utf-8 -*-

import six
import copy
import logging
import warnings
from functools import wraps

from . import signals
from . import exceptions
from .fields import Field, ListField, ForeignList, AbstractForeignList
from .storage import Storage
from .query import QueryBase, RawQuery, QueryGroup
from .frozen import FrozenDict
from .cache import Cache
from .writequeue import WriteQueue, WriteAction


logger = logging.getLogger(__name__)


class ContextLogger(object):

    @staticmethod
    def sort_func(e):
        return (e.xtra._name if e.xtra else None, e.func.__name__)

    def report(self, sort_func=None):
        return self.logger.report(sort_func or self.sort_func)

    def __init__(self, log_level=None, xtra=None, sort_func=None):
        self.log_level = log_level
        self.xtra = xtra
        self.sort_func = sort_func or self.sort_func
        self.logger = Storage.logger

    def __enter__(self):
        self.listening = self.logger.listen(self.xtra)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.listening:
            report = self.logger.report(
                lambda e: (e.xtra._name if e.xtra else None, e.func.__name__)
            )
            if self.log_level is not None:
                logging.log(self.log_level, report)
            self.logger.clear()
        self.logger.pop()


def deref(data, keys, missing=None):
    if keys[0] in data:
        if len(keys) == 1:
            return data[keys[0]]
        return deref(data[keys[0]], keys[1:], missing=missing)
    return missing


def flatten_backrefs(data, stack=None):

    stack = stack or []

    if isinstance(data, list):
        return [(stack, item) for item in data]

    out = []
    for key, val in data.items():
        out.extend(flatten_backrefs(val, stack + [key]))

    return out


def log_storage(func):

    @wraps(func)
    def wrapped(this, *args, **kwargs):

        cls = this if isinstance(this, type) else type(this)

        with ContextLogger(log_level=this._log_level, xtra=cls):
            return func(this, *args, **kwargs)

    return wrapped


def warn_if_detached(func):
    """ Warn if self / cls is detached. """
    @wraps(func)
    def wrapped(this, *args, **kwargs):
        # Check for _detached in __dict__ instead of using hasattr
        # to avoid infinite loop in __getattr__
        if '_detached' in this.__dict__ and this._detached:
            warnings.warn('here')
        return func(this, *args, **kwargs)
    return wrapped


def has_storage(func):
    """ Ensure that self/cls contains a Storage backend. """
    @wraps(func)
    def wrapped(*args, **kwargs):
        me = args[0]
        if not hasattr(me, '_storage') or \
                not me._storage:
            raise exceptions.ImproperConfigurationError(
                'No storage backend attached to schema <{0}>.'
                    .format(me._name.upper())
            )
        return func(*args, **kwargs)
    return wrapped


class ObjectMeta(type):

    def add_field(cls, name, field):

        # Skip if not descriptor
        if not isinstance(field, Field):
            return

        # Memorize parent references
        field._schema_class = cls
        field._field_name = name

        # Check for primary key
        if field._is_primary:
            if cls._primary_name is None:
                cls._primary_name = name
                cls._primary_type = field.data_type
            else:
                raise AttributeError(
                    'Multiple primary keys are not supported.')

        # Wrap in list
        if field._list and not isinstance(field, ListField):
            field = ListField(
                field,
                **field._kwargs
            )
            # Memorize parent references
            field._schema_class = cls
            field._field_name = name
            # Set parent pointer of child field to list field
            field._field_instance._list_container = field

        # Subscribe to schema events
        field.subscribe(sender=cls)

        # Store descriptor to cls, cls._fields
        setattr(cls, name, field)
        cls._fields[name] = field

    def __init__(cls, name, bases, dct):

        # Run super-metaclass __init__
        super(ObjectMeta, cls).__init__(name, bases, dct)

        # Store prettified name
        cls._name = name.lower()

        # Store parameters from _meta
        my_meta = cls.__dict__.get('_meta', {})

        cls._is_optimistic = my_meta.get('optimistic', False)
        cls._is_abstract = my_meta.get('abstract', False)
        cls._log_level = my_meta.get('log_level', None)
        cls._version_of = my_meta.get('version_of', None)
        cls._version = my_meta.get('version', 1)
        cls._record_validators = my_meta.get('validators', [])

        # Prepare fields
        cls._fields = {}
        cls._primary_name = None
        cls._primary_type = None

        for key, value in cls.__dict__.items():
            cls.add_field(key, value)

        for base in bases:
            if not hasattr(base, '_fields') or not isinstance(base._fields, dict):
                continue
            for key, value in base._fields.items():
                cls.add_field(key, copy.deepcopy(value))

        # Impute field named _id as primary if no primary field specified;
        # must be exactly one primary field unless abstract
        if cls._fields:
            cls._is_root = False
            if cls._primary_name is None:
                if '_id' in cls._fields:
                    primary_field = cls._fields['_id']
                    primary_field._is_primary = True
                    if 'index' not in primary_field._kwargs or not primary_field._kwargs['index']:
                        primary_field._index = True
                    cls._primary_name = '_id'
                    cls._primary_type = cls._fields['_id'].data_type
                elif not cls._is_abstract:
                    raise AttributeError(
                        'Schemas must either define a field named _id or '
                        'specify exactly one field as primary.')
            # Register
            cls.register_collection()
        else:
            cls._is_root = True

    @property
    def _translator(cls):
        return cls._storage[0].translator


@six.add_metaclass(ObjectMeta)
class StoredObject(object):
    """
    Base class to be used for models.
    """

    _collections = {}

    _cache = Cache()
    _object_cache = Cache()
    queue = WriteQueue()

    def __init__(self, **kwargs):

        # Crash if abstract
        if self._is_abstract:
            raise TypeError('Cannot instantiate abstract schema')

        self.__backrefs = {}
        self._dirty = False
        self._detached = False
        self._is_loaded = kwargs.pop('_is_loaded', False)
        self._stored_key = None

        # Impute non-lazy default values (e.g. datetime with auto_now=True)
        for value in self._fields.values():
            if not value.lazy_default:
                value.__set__(self, value._gen_default(), safe=True)

        # Add kwargs to instance
        for key, value in kwargs.items():
            try:
                field = self._fields[key]
                field.__set__(self, value, safe=True)
            except KeyError:
                if key == '__backrefs':
                    key = '_StoredObject__backrefs'
                setattr(self, key, value)

        if self._is_loaded:
            self._set_cache(self._primary_key, self, kwargs)

    def __eq__(self, other):
        try:
            if self is other:
                return True
            if self._primary_key != other._primary_key:
                return False
            return self.to_storage() == other.to_storage()
        except (AttributeError, TypeError):
            # Can't compare with "other". Try the reverse comparison
            return NotImplemented

    def __ne__(self, other):
        equal = self.__eq__(other)
        return equal if equal is NotImplemented else not equal

    def __hash__(self):
        # TODO: Is this the right thing to do?
        return id(self)

    @warn_if_detached
    def __unicode__(self):
        return unicode({field : unicode(getattr(self, field)) for field in self._fields})

    @warn_if_detached
    def __str__(self):
        return unicode(self).decode('ascii', 'replace')

    @classmethod
    def register_collection(cls):
        cls._collections[cls._name] = cls

    @classmethod
    def get_collection(cls, name):
        return cls._collections[name.lower()]

    @property
    def _primary_key(self):
        return getattr(self, self._primary_name)

    @_primary_key.setter
    def _primary_key(self, value):
        setattr(self, self._primary_name, value)

    @property
    def _storage_key(self):
        """ Primary key passed through translator. """
        return self._pk_to_storage(self._primary_key)

    @property
    @has_storage
    def _translator(self):
        return self.__class__._translator

    @has_storage
    def to_storage(self, translator=None, clone=False):

        data = {}

        for field_name, field_object in self._fields.items():

            # Ignore primary and foreign fields if cloning
            # TODO: test this
            if clone:
                if field_object._is_primary or field_object._is_foreign:
                    continue
            field_value = field_object.to_storage(
                field_object._get_underlying_data(self),
                translator
            )
            data[field_name] = field_value

        data['_version'] = self._version
        if not clone and self.__backrefs:
            data['__backrefs'] = self.__backrefs

        return data

    @classmethod
    @has_storage
    def from_storage(cls, data, translator=None):

        result = {}

        for key, value in data.items():

            field_object = cls._fields.get(key, None)

            if isinstance(field_object, Field):
                data_value = data[key]
                if data_value is None:
                    value = None
                    result[key] = None
                else:
                    value = field_object.from_storage(data_value, translator)
                result[key] = value

            else:
                result[key] = value

        return result

    def clone(self):
        return self.load(
            data=self.to_storage(clone=True),
            _is_loaded=False
        )

    # Backreferences

    @property
    def _backrefs(self):
        return FrozenDict(**self.__backrefs)

    @_backrefs.setter
    def _backrefs(self, _):
        raise exceptions.ModularOdmException('Cannot modify _backrefs.')

    @property
    def _backrefs_flat(self):
        return flatten_backrefs(self.__backrefs)

    def _remove_backref(self, backref_key, parent, parent_field_name, strict=False):
        try:
            self.__backrefs[backref_key][parent._name][parent_field_name].remove(parent._primary_key)
            self.save(force=True)
        except (KeyError, ValueError):
            if strict:
                raise

    def _update_backref(self, backref_key, parent, parent_field_name):
        updated = False
        try:
            refs = self.__backrefs[backref_key][parent._name][parent_field_name]
            if refs[refs.index(parent._stored_key)] != parent._primary_key:
                refs[refs.index(parent._stored_key)] = parent._primary_key
                updated = True
        except (KeyError, ValueError):
            self._set_backref(backref_key, parent_field_name, parent)
            return True
        if updated:
            self.save(force=True)
            return True
        return False

    def _set_backref(self, backref_key, parent_field_name, backref_value):

        backref_value_class_name = backref_value.__class__._name
        backref_value_primary_key = backref_value._primary_key

        if backref_value_primary_key is None:
            raise exceptions.DatabaseError('backref object\'s primary key must be saved first')

        if backref_key not in self.__backrefs:
            self.__backrefs[backref_key] = {}
        if backref_value_class_name not in self.__backrefs[backref_key]:
            self.__backrefs[backref_key][backref_value_class_name] = {}
        if parent_field_name not in self.__backrefs[backref_key][backref_value_class_name]:
            self.__backrefs[backref_key][backref_value_class_name][parent_field_name] = []

        append_to = self.__backrefs[backref_key][backref_value_class_name][parent_field_name]
        if backref_value_primary_key not in append_to:
            append_to.append(backref_value_primary_key)

        self.save(force=True)

    @classmethod
    def set_storage(cls, storage):

        if not isinstance(storage, Storage):
            raise TypeError('Argument to set_storage must be an instance of Storage.')
        if not hasattr(cls, '_storage'):
            cls._storage = []

        for field_name, field_object in cls._fields.items():
            if field_object._index:
                storage._ensure_index(field_name)

        cls._storage.append(storage)

    # Caching ################################################################

    @classmethod
    def _is_cached(cls, key):
        return cls._object_cache.get(cls._name, key) is not None

    @classmethod
    def _load_from_cache(cls, key):
        trans_key = cls._pk_to_storage(key)
        return cls._object_cache.get(cls._name, trans_key)

    @classmethod
    def _set_cache(cls, key, obj, data=None):
        trans_key = cls._pk_to_storage(key)
        cls._object_cache.set(cls._name, trans_key, obj)
        cls._cache.set(cls._name, trans_key, data)

    @classmethod
    def _get_cache(cls, key):
        trans_key = cls._pk_to_storage(key)
        return cls._object_cache.get(cls._name, trans_key)

    @classmethod
    def _get_cached_data(cls, key):
        return cls._cache.get(cls._name, key)

    def get_changed_fields(self, cached_data, storage_data):
        """Get fields that differ between the cache_sandbox and the current object.
        Validation and after_save methods should only be run on diffed
        fields.

        :param cached_data: Storage-formatted data from cache_sandbox
        :param storage_data: Storage-formatted data from object
        :return: List of diffed fields

        """
        if not self._is_loaded or cached_data is None:
            return []

        return [
            field
            for field in self._fields
            if cached_data.get(field) != storage_data.get(field)
        ]

    # Cache clearing

    @classmethod
    def _clear_data_cache(cls, key=None):
        if not cls._fields:
            cls._cache.clear()
        elif key is not None:
            cls._cache.pop(cls._name, key)
        else:
            cls._cache.clear_schema(cls._name)

    @classmethod
    def _clear_object_cache(cls, key=None):
        if not cls._fields:
            cls._object_cache.clear()
        elif key is not None:
            cls._object_cache.pop(cls._name, key)
        else:
            cls._object_cache.clear_schema(cls._name)

    @classmethod
    def _clear_caches(cls, key=None):
        cls._clear_data_cache(key)
        cls._clear_object_cache(key)

    ###########################################################################

    @classmethod
    def _to_primary_key(cls, value):

        if value is None:
            return value

        if isinstance(value, cls):
            return value._primary_key

        return cls._check_pk_type(value)

    @classmethod
    def _check_pk_type(cls, key):

        if isinstance(key, cls._primary_type):
            return key

        try:
            cls._primary_type()
            cast_type = cls._primary_type
        except:
            cast_type = str

        try:
            key = cast_type(key)
        except:
            raise TypeError(
                'Invalid key type: {key}, {type}, {ptype}.'.format(
                    key=key, type=type(key), ptype=cast_type
                )
            )

        return key

    @classmethod
    @has_storage
    @log_storage
    def load(cls, key=None, data=None, _is_loaded=True):
        """Get a record by its primary key.
        """
        # Emit load signal
        signals.load.send(
            cls,
            key=key,
            data=data,
        )

        if key is not None:
            key = cls._check_pk_type(key)
            cached_object = cls._load_from_cache(key)
            if cached_object is not None:
                return cached_object

        # Try loading from backend
        if data is None:
            data = cls._storage[0].get(cls._primary_name, cls._pk_to_storage(key))

        # if not found, return None
        if data is None:
            return None

        # Convert storage data to ODM
        data = cls.from_storage(data)

        if cls._version_of and '_version' in data and data['_version'] != cls._version:
            old_object = cls._version_of.load(data=data)
            new_object = cls(_is_loaded=_is_loaded)

            cls.migrate(old_object, new_object)
            new_object._stored_key = new_object._primary_key

            return new_object

        ret = cls(_is_loaded=_is_loaded, **data)
        ret._stored_key = ret._primary_key

        return ret

    @classmethod
    def migrate_all(cls):
        """Migrate all records in this collection."""
        for record in cls.find():
            record.save()

    @classmethod
    def migrate(cls, old, new, verbose=True, dry_run=False, rm_refs=True):
        """Migrate record to new schema.

        :param old: Record from original schema
        :param new: Record from new schema
        :param verbose: Print detailed info
        :param dry_run: Dry run; make no changes if true
        :param rm_refs: Remove references on deleted fields

        """
        if verbose:
            logging.basicConfig(format='%(levelname)s %(filename)s: %(message)s',
                                level=logging.DEBUG)
        # Check deleted, added fields
        deleted_fields = [field for field in old._fields if field not in new._fields]
        added_fields = [field for field in new._fields if field not in old._fields]
        logging.info('Will delete fields: {0}'.format(deleted_fields))
        logging.info('Will add fields: {0}'.format(added_fields))

        # Check change in primary key
        if old._primary_name != new._primary_name:
            logging.info("The primary key will change from {old_name}: {old_field} to "
                "{new_name}: {new_field} in this migration. Primary keys and "
                "backreferences will not be automatically migrated. If you want "
                "to migrate primary keys, you should handle this in your "
                "migrate() method."
                    .format(old_name=old._primary_name,
                            old_field=old._fields[old._primary_name],
                            new_name=new._primary_name,
                            new_field=new._fields[new._primary_name]))

        # Copy fields to new object
        for field in old._fields:

            # Delete forward references on deleted fields
            if field not in cls._fields:
                if rm_refs:
                    logging.info("Backreferences to this object keyed on foreign "
                        "field {name}: {field} will be deleted in this migration. "
                        "To prevent this behavior, re-run with <rm_fwd_refs> "
                        "set to False.".format(name=field,
                                              field=old._fields[field]))
                    if not dry_run:
                        rm_fwd_refs(old)
                else:
                    logging.info("Backreferences to this object keyed on foreign field "
                        "{name}: {field} will be not deleted in this migration. "
                        "To add this behavior, re-run with <rm_fwd_refs> "
                        "set to True.".format(name=field,
                                            field=old._fields[field]))
                continue

            # Check for field change
            old_field_obj = old._fields[field]
            new_field_obj = new._fields[field]
            if old_field_obj != new_field_obj:
                if not old_field_obj._required and new_field_obj._required:
                    logging.info("Field {name!r} is now required "
                            "and therefore needs a default value "
                            "for existing records. You can set "
                            "this value in the _migrate() method. "
                            "\nExample: "
                            "\n    if not old.{name}:"
                            "\n        new.{name} = 'default value'"
                            .format(name=field))
                else:
                    logging.info(
                        "Old field {name}: {old_field} differs from new field "
                        "{name}: {new_field}. This field will not be "
                        "automatically migrated. If you want to migrate this "
                        "field, you should handle this in your migrate() "
                        "method.".format(name=field,
                                         old_field=old_field_obj,
                                         new_field=new_field_obj)
                    )
                continue

            # Copy values of retained fields
            if not dry_run:
                field_object = cls._fields[field]
                field_object.__set__(
                    new,
                    getattr(old, field),
                    safe=True
                )

        # Copy backreferences
        if not dry_run:
            new.__backrefs = old.__backrefs

        # Run custom migration
        if not dry_run:
            cls._migrate(old, new)

    @classmethod
    def _migrate(cls, old, new):
        """Subclasses can override this class to perform a custom migration.
        This is run after the migrate() method.

        Example:
        ::

            class NewSchema(StoredObject):
                _id = fields.StringField(primary=True, index=True)
                my_string = fields.StringField()

                @classmethod
                def _migrate(cls, old, new):
                    new.my_string = old.my_string + 'yo'

                _meta = {
                    'version_of': OldSchema,
                    'version': 2,
                    'optimistic': True
                }

        :param old: Record from original schema
        :param new: Record from new schema
        """
        return new

    @classmethod
    def explain_migration(cls):
        logging.basicConfig(format='%(levelname)s %(filename)s: %(message)s',
                            level=logging.DEBUG)

        classes = [cls]
        methods = [cls._migrate]
        klass = cls
        while klass._version and klass._version_of:
            classes.insert(0, klass._version_of)
            try:
                methods.insert(0, klass._migrate)
            except AttributeError:
                methods.insert(0, None)
            klass = klass._version_of

        for step in range(len(classes) - 1):

            fr = classes[step]
            to = classes[step + 1]

            logging.info('From schema {0}'.format(fr._name))
            logging.info('\n'.join('\t{0}'.format(field) for field in fr._fields))
            logging.info('')

            logging.info('To schema {0}'.format(to._name))
            logging.info('\n'.join('\t{0}'.format(field) for field in to._fields))
            logging.info('')

            to.migrate(fr, to, verbose=True, dry_run=True)

    @classmethod
    def _must_be_loaded(cls, value):
        if value is not None and not value._is_loaded:
            raise exceptions.DatabaseError('Record must be loaded.')

    @has_storage
    @log_storage
    def _optimistic_insert(self):
        self._primary_key = self._storage[0]._optimistic_insert(
            self._primary_name,
            self.to_storage()
        )

    def validate_record(self):
        """Apply record-level validation. Run on `save`.

        """
        for validator in self._record_validators:
            validator(self)

    @has_storage
    @log_storage
    def save(self, force=False):
        """Save a record.

        :param bool force: Save even if no fields have changed; used to update
            back-references
        :returns: List of changed fields
        """
        if self._detached:
            raise exceptions.DatabaseError('Cannot save detached object.')

        for field_name, field_object in self._fields.items():
            if hasattr(field_object, 'on_before_save'):
                field_object.on_before_save(self)

        signals.before_save.send(
            self.__class__,
            instance=self
        )

        cached_data = self._get_cached_data(self._stored_key)
        storage_data = self.to_storage()

        if self._primary_key is not None and cached_data is not None:
            fields_changed = set(
                self.get_changed_fields(cached_data, storage_data)
            )
        else:
            fields_changed = set(self._fields.keys())

        # Quit if no diffs
        if not fields_changed and not force:
            return []

        # Apply field-level validation
        for field_name in fields_changed:
            field_object = self._fields[field_name]
            field_object.do_validate(getattr(self, field_name), self)

        # Apply record-level validation
        self.validate_record()

        primary_changed = (
            self._primary_key != self._stored_key
            and
            self._primary_name in fields_changed
        )

        if self._is_loaded:
            if primary_changed and not getattr(self, '_updating_key', False):
                self.delegate(
                    self._storage[0].remove,
                    False,
                    RawQuery(self._primary_name, 'eq', self._stored_key)
                )
                self._clear_caches(self._stored_key)
                self.insert(self._primary_key, storage_data)
            else:
                self.update_one(self, storage_data=storage_data, saved=True, inmem=True)
        elif self._is_optimistic and self._primary_key is None:
            self._optimistic_insert()
        else:
            self.insert(self._primary_key, storage_data)

        # if primary key has changed, follow back references and update
        # AND
        # run after_save or after_save_on_difference

        if self._is_loaded and primary_changed:
            if not getattr(self, '_updating_key', False):
                self._updating_key = True
                update_backref_keys(self)
                self._stored_key = self._primary_key
                self._updating_key = False
        else:
            self._stored_key = self._primary_key

        self._is_loaded = True

        signals.save.send(
            self.__class__,
            instance=self,
            fields_changed=fields_changed,
            cached_data=cached_data or {},
        )

        storage_data[self._primary_name] = self._storage_key
        self._set_cache(self._primary_key, self, storage_data)

        return fields_changed

    def update_fields(self, **kwargs):
        """Update multiple fields, specified by keyword arguments.

        Example::

            person.update(given_name='Fred', family_name='Mercury')

        ... is equivalent to ... ::


            person.given_name = 'Fred'
            person.family_name = 'Mercury'

        :param **kwargs: field names and the values to set
        """
        for key, value in kwargs.items():
            self._fields[key].__set__(self, value, safe=True)


    def reload(self):

        storage_data = self._storage[0].get(self._primary_name, self._storage_key)

        for key, value in storage_data.items():
            field_object = self._fields.get(key, None)
            if isinstance(field_object, Field):
                data_value = storage_data[key]
                if data_value is None:
                    value = None
                else:
                    value = field_object.from_storage(data_value)
                field_object.__set__(self, value, safe=True)
            elif key == '__backrefs':
                self._StoredObject__backrefs = value

        self._stored_key = self._primary_key
        self._set_cache(self._storage_key, self, storage_data)

    @warn_if_detached
    def __getattr__(self, item):

        errmsg = '{cls} object has no attribute {item}'.format(
            cls=self.__class__.__name__,
            item=item
        )

        if item in self.__backrefs:
            backrefs = []
            for parent, rest0 in six.iteritems(self.__backrefs[item]):
                for field, rest1 in six.iteritems(rest0):
                    backrefs.extend([
                        (key, parent)
                        for key in rest1
                    ])
            return AbstractForeignList(backrefs)

        # Retrieve back-references
        if '__' in item and not item.startswith('__'):
            item_split = item.split('__')
            if len(item_split) == 2:
                parent_schema_name, backref_key = item_split
                backrefs = deref(self.__backrefs, [backref_key, parent_schema_name], missing={})
                ids = sum(
                    backrefs.values(),
                    []
                )
            elif len(item_split) == 3:
                parent_schema_name, backref_key, parent_field_name = item_split
                ids = deref(self.__backrefs, [backref_key, parent_schema_name, parent_field_name], missing=[])
            else:
                raise AttributeError(errmsg)
            try:
                base_class = self.get_collection(parent_schema_name)
            except KeyError:
                raise exceptions.ModularOdmException(
                    'Unknown schema <{0}>'.format(
                        parent_schema_name
                    )
                )
            return ForeignList(ids, literal=True, base_class=base_class)

        raise AttributeError(errmsg)

    @warn_if_detached
    def __setattr__(self, key, value):
        if key not in self._fields and not key.startswith('_'):
            warnings.warn('Setting an attribute that is neither a field nor a protected value.')
        super(StoredObject, self).__setattr__(key, value)

    # Querying ######

    @classmethod
    def _parse_key_value(cls, value):
        if isinstance(value, StoredObject):
            return value._primary_key, value
        return value, cls.load(cls._pk_to_storage(value))

    @classmethod
    @has_storage
    def _pk_to_storage(cls, key):
        return cls._fields[cls._primary_name].to_storage(key)

    @classmethod
    def _process_query(cls, query):

        if isinstance(query, RawQuery):
            field = cls._fields.get(query.attribute)
            if field is None:
                return
            if field._is_foreign:
                if getattr(query.argument, '_fields', None):
                    if field._is_abstract:
                        query.argument = (
                            query.argument._primary_key,
                            query.argument._name,
                        )
                    else:
                        query.argument = query.argument._primary_key
        elif isinstance(query, QueryGroup):
            for node in query.nodes:
                cls._process_query(node)

    @classmethod
    @has_storage
    @log_storage
    def find(cls, query=None, **kwargs):
        """

        :param query:
        :param kwargs:
        :return: an iterable of :class:`StoredObject` instances
        """
        cls._process_query(query)
        return cls._storage[0].QuerySet(
            cls,
            cls._storage[0].find(query, **kwargs)
        )

    @classmethod
    @has_storage
    @log_storage
    def find_one(cls, query=None, **kwargs):
        cls._process_query(query)
        stored_data = cls._storage[0].find_one(query, **kwargs)
        return cls.load(
            key=stored_data[cls._primary_name],
            data=stored_data
        )

    # Queueing

    @classmethod
    def delegate(cls, method, conflict=None, *args, **kwargs):
        """Execute or queue a database action. Variable positional and keyword
        arguments are passed to the provided method.

        :param function method: Method to execute or queue
        :param bool conflict: Potential conflict between cache_sandbox and backend,
            e.g., in the event of bulk updates or removes that bypass the
            cache_sandbox

        """
        if cls.queue.active:
            action = WriteAction(method, *args, **kwargs)
            if conflict:
                logger.warn('Delayed write {0!r} may cause the cache to '
                            'diverge from the database until changes are '
                            'committed.'.format(action))
            cls.queue.push(action)
        else:
            method(*args, **kwargs)

    @classmethod
    def start_queue(cls):
        """Start the queue. Between calling `start_queue` and `commit_queue`,
        all writes will be deferred to the queue.

        """
        cls.queue.start()

    @classmethod
    def clear_queue(cls):
        """Clear the queue.

        """
        cls.queue.clear()

    @classmethod
    def cancel_queue(cls):
        """Cancel any pending actions. This method clears the queue and also
        clears caches if any actions are pending.

        """
        if cls.queue:
            cls._cache.clear()
            cls._object_cache.clear()
        cls.clear_queue()

    @classmethod
    def commit_queue(cls):
        """Commit all queued actions. If any actions fail, clear caches. Note:
        the queue will be cleared whether an error is raised or not.

        """
        try:
            cls.queue.commit()
            cls.clear_queue()
        except:
            cls.cancel_queue()
            raise

    @classmethod
    def subscribe(cls, signal_name, weak=True):
        """

        :param str signal_name: Name of signal to subscribe to; must be found
            in ``signals.py``.
        :param bool weak: Create weak reference to callback
        :returns: Decorator created by ``Signal::connect_via``
        :raises: ValueError if signal is not found

        Example usage: ::

            >>> @Schema.subscribe('before_save')
            ... def listener(cls, instance):
            ...     instance.value += 1

        """
        try:
            signal = getattr(signals, signal_name)
        except AttributeError:
            raise ValueError(
                'Signal {0} not found'.format(signal_name)
            )
        sender = None if cls._is_root else cls
        return signal.connect_via(sender, weak)

    @classmethod
    @has_storage
    def insert(cls, key, val):
        cls.delegate(
            cls._storage[0].insert,
            False,
            cls._primary_name,
            cls._pk_to_storage(key),
            val
        )

    @classmethod
    def _includes_foreign(cls, keys):
        for key in keys:
            if key in cls._fields and cls._fields[key]._is_foreign:
                return True
        return False

    @classmethod
    def _data_to_storage(cls, data):

        storage_data = {}

        for key, value in data.items():
            if key in cls._fields:
                field_object = cls._fields[key]
                if key == cls._primary_name:
                    continue
                storage_data[key] = field_object.to_storage(value)
            else:
                storage_data[key] = value

        return storage_data

    def _update_in_memory(self, storage_data):
        for field_name, data_value in storage_data.items():
            field_object = self._fields[field_name]
            field_object.__set__(self, data_value, safe=True)
        self.save()

    @classmethod
    def _which_to_obj(cls, which):
        if isinstance(which, QueryBase):
            return cls.find_one(which)
        if isinstance(which, StoredObject):
            return which
        return cls.load(cls._pk_to_storage(which))

    @classmethod
    @has_storage
    def update_one(cls, which, data=None, storage_data=None, saved=False, inmem=False):

        storage_data = storage_data or cls._data_to_storage(data)
        includes_foreign = cls._includes_foreign(storage_data.keys())
        obj = cls._which_to_obj(which)

        if saved or not includes_foreign:
            cls.delegate(
                cls._storage[0].update,
                False,
                RawQuery(
                    cls._primary_name, 'eq', obj._primary_key
                ),
                storage_data,
            )
            if obj and not inmem:
                obj._dirty = True
            if not saved:
                cls._clear_caches(obj._storage_key)
        else:
            obj._update_in_memory(storage_data)

    @classmethod
    @has_storage
    def update(cls, query, data=None, storage_data=None):

        storage_data = storage_data or cls._data_to_storage(data)
        includes_foreign = cls._includes_foreign(storage_data.keys())

        objs = cls.find(query)
        keys = objs.get_keys()

        if not includes_foreign:
            cls.delegate(
                cls._storage[0].update,
                True,
                query,
                storage_data
            )
            for key in keys:
                obj = cls._get_cache(key)
                if obj is not None:
                    obj._dirty = True

        else:
            for obj in objs:
                obj._update_in_memory(storage_data)

    @classmethod
    @has_storage
    def remove_one(cls, which, rm=True):
        """Remove an object, along with its references and back-references.
        Remove the object from the cache_sandbox and sets its _detached flag to True.

        :param which: Object selector: Query, StoredObject, or primary key
        :param rm: Remove data from backend

        """
        # Look up object
        obj = cls._which_to_obj(which)

        # Remove references
        rm_fwd_refs(obj)
        rm_back_refs(obj)

        # Remove from cache_sandbox
        cls._clear_caches(obj._storage_key)

        # Remove from backend
        if rm:
            cls.delegate(
                cls._storage[0].remove,
                False,
                RawQuery(obj._primary_name, 'eq', obj._storage_key)
            )

        # Set detached
        obj._detached = True

    @classmethod
    @has_storage
    def remove(cls, query=None):
        """Remove objects by query.

        :param query: Query object

        """
        objs = cls.find(query)

        for obj in objs:
            cls.remove_one(obj, rm=False)

        cls.delegate(
            cls._storage[0].remove,
            False,
            query
        )

def rm_fwd_refs(obj):
    """When removing an object, other objects with references to the current
    object should remove those references. This function identifies objects
    with forward references to the current object, then removes those
    references.

    :param obj: Object to which forward references should be removed

    """
    for stack, key in obj._backrefs_flat:

        # Unpack stack
        backref_key, parent_schema_name, parent_field_name = stack

        # Get parent info
        parent_schema = obj._collections[parent_schema_name]
        parent_key_store = parent_schema._pk_to_storage(key)
        parent_object = parent_schema.load(parent_key_store)
        if parent_object is None:
            continue

        # Remove forward references
        if parent_object._fields[parent_field_name]._list:
            getattr(parent_object, parent_field_name).remove(obj)
        else:
            parent_field_object = parent_object._fields[parent_field_name]
            setattr(parent_object, parent_field_name, parent_field_object._gen_default())

        # Save
        parent_object.save()

def _collect_refs(obj, fields=None):
    """

    """
    refs = []
    fields = fields or []

    for field_name, field_object in obj._fields.items():

        # Skip if not foreign field
        if not field_object._is_foreign:
            continue

        # Skip if value is None
        value = getattr(obj, field_name)
        if value is None:
            continue

        # Skip if not in fields
        if fields and field_name not in fields:
            continue

        field_refs = []

        # Build list of linked objects if ListField, else single field
        if isinstance(field_object, ListField):
            field_refs.extend([v for v in value if v])
            field_instance = field_object._field_instance
        else:
            field_refs.append(value)
            field_instance = field_object

        # Skip if field does not specify back-references
        if not field_instance._backref_field_name:
            continue

        refs.extend([
            {
                'value': ref,
                'field_name': field_name,
                'field_instance': field_instance,
            }
            for ref in field_refs
        ])

    return refs

def rm_back_refs(obj):
    """When removing an object with foreign fields, back-references from
    other objects to the current object should be deleted. This function
    identifies foreign fields of the specified object whose values are not
    None and which specify back-reference keys, then removes back-references
    from linked objects to the specified object.

    :param obj: Object for which back-references should be removed

    """
    for ref in _collect_refs(obj):
        ref['value']._remove_backref(
            ref['field_instance']._backref_field_name,
            obj,
            ref['field_name'],
            strict=False
        )

def ensure_backrefs(obj, fields=None):
    """Ensure that all forward references on the provided object have the
    appropriate backreferences.

    :param StoredObject obj: Database record
    :param list fields: Optional list of field names to check

    """
    for ref in _collect_refs(obj, fields):
        updated = ref['value']._update_backref(
            ref['field_instance']._backref_field_name,
            obj,
            ref['field_name'],
        )
        if updated:
            logging.debug('Updated reference {}:{}:{}:{}:{}'.format(
                obj._name, obj._primary_key, ref['field_name'],
                ref['value']._name, ref['value']._primary_key,
            ))

def update_backref_keys(obj):
    """

    """
    for ref in _collect_refs(obj):
        ref['value']._update_backref(
            ref['field_instance']._backref_field_name,
            obj,
            ref['field_name'],
        )

    for stack, key in obj._backrefs_flat:

        # Unpack stack
        backref_key, parent_schema_name, parent_field_name = stack

        # Get parent info
        parent_schema = obj._collections[parent_schema_name]
        parent_key_store = parent_schema._pk_to_storage(key)
        parent_object = parent_schema.load(parent_key_store)
        if parent_object is None:
            continue

        #
        field_object = parent_object._fields[parent_field_name]
        if field_object._list:
            value = getattr(parent_object, parent_field_name)
            if field_object._is_abstract:
                idx = value.index((obj._stored_key, obj._name))
                value[idx] = (obj._primary_key, obj._name)
            else:
                idx = value.index(obj._stored_key)
                value[idx] = obj
        else:
            setattr(parent_object, parent_field_name, obj)

        # Save
        parent_object.save()



import six
from six.moves.urllib_parse import urlsplit, urlunsplit

from modularodm.exceptions import (
    ValidationError,
    ValidationTypeError,
    ValidationValueError,
)

from bson import ObjectId

class Validator(object):

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

class TypeValidator(Validator):

    def _as_list(self, value):

        if isinstance(value, (tuple, list)):
            return value
        return [value]

    def __init__(self, allowed_types, forbidden_types=None):
        self.allowed_types = self._as_list(allowed_types)
        self.forbidden_types = self._as_list(forbidden_types) if forbidden_types else []

    def __call__(self, value):

        for ftype in self.forbidden_types:
            if isinstance(value, ftype):
                self._raise(value)

        for atype in self.allowed_types:
            if isinstance(value, atype):
                return

        self._raise(value)

    def _raise(self, value):

        raise ValidationTypeError(
            'Received invalid value {} of type {}'.format(
                value, type(value)
            )
        )

validate_string = TypeValidator(six.string_types)
validate_integer = TypeValidator(
    allowed_types=int,
    forbidden_types=bool
)
validate_float = TypeValidator(float)
validate_boolean = TypeValidator(bool)
validate_objectid = TypeValidator(ObjectId)

from ..fields.lists import List
validate_list = TypeValidator(List)

import datetime
validate_datetime = TypeValidator(datetime.datetime)

# Adapted from Django RegexValidator
import re
class RegexValidator(Validator):

    def __init__(self, regex=None, flags=0):

        if regex is not None:
            self.regex = re.compile(regex, flags=flags)

    def __call__(self, value):

        if not self.regex.search(value):
            raise ValidationError(
                'Value must match regex {0} and flags {1}; received value <{2}>'.format(
                    self.regex.pattern,
                    self.regex.flags,
                    value
                )
            )

# Adapted from Django URLValidator
class URLValidator(RegexValidator):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    # message = _('Enter a valid URL.')

    def __call__(self, value):
        try:
            super(URLValidator, self).__call__(value)
        except ValidationError as e:
            # Trivial case failed. Try for possible IDN domain
            if value:
                # value = force_text(value)
                scheme, netloc, path, query, fragment = urlsplit(value)
                try:
                    netloc = netloc.encode('idna').decode('ascii')  # IDN -> ACE
                except UnicodeError:  # invalid domain part
                    raise e
                url = urlunsplit((scheme, netloc, path, query, fragment))
                super(URLValidator, self).__call__(url)
            else:
                raise
        else:
            pass
            # url = value


class BaseValidator(Validator):

    compare = lambda self, a, b: a is not b
    clean = lambda self, x: x
    message = 'Ensure this value is %(limit_value)s (it is %(show_value)s).'
    code = 'limit_value'

    def __init__(self, limit_value):
        self.limit_value = limit_value

    def __call__(self, value):
        cleaned = self.clean(value)
        params = {'limit_value': self.limit_value, 'show_value': cleaned}
        if self.compare(cleaned, self.limit_value):
            raise ValidationValueError(self.message.format(**params))


class MaxValueValidator(BaseValidator):

    compare = lambda self, a, b: a > b
    message = 'Ensure this value is less than or equal to {limit_value}.'
    code = 'max_value'


class MinValueValidator(BaseValidator):

    compare = lambda self, a, b: a < b
    message = 'Ensure this value is greater than or equal to {limit_value}.'
    code = 'min_value'


class MinLengthValidator(BaseValidator):

    compare = lambda self, a, b: a < b
    clean = lambda self, x: len(x)
    message = 'Ensure this value has length of at least {limit_value} (it has length {show_value}).'
    code = 'min_length'


class MaxLengthValidator(BaseValidator):

    compare = lambda self, a, b: a > b
    clean = lambda self, x: len(x)
    message = 'Ensure this value has length of at most {limit_value} (it has length {show_value}).'
    code = 'max_length'




class QueryBase(object):

    def __or__(self, other):
        return QueryGroup('or', self, other)

    def __and__(self, other):
        return QueryGroup('and', self, other)

    def __invert__(self):
        return QueryGroup('not', self)

class QueryGroup(QueryBase):

    def __init__(self, operator, *args):

        self.operator = operator

        self.nodes = []
        for node in args:
            if not isinstance(node, QueryBase):
                raise TypeError('Nodes must be Query objects.')
            if isinstance(node, QueryGroup) and node.operator == operator:
                self.nodes += node.nodes
            else:
                self.nodes.append(node)

    def __repr__(self):

        return '{0}({1})'.format(
            self.operator.upper(),
            ', '.join(repr(node) for node in self.nodes)
        )

class RawQuery(QueryBase):

    def __init__(self, attribute, operator, argument):

        self.attribute = attribute
        self.operator = operator
        self.argument = argument

    def __repr__(self):

        return 'RawQuery({}, {}, {})'.format(
            self.attribute,
            self.operator,
            self.argument
        )
